#! /usr/bin/env python
# -*- coding: utf-8 -*-

# lexiko.py - Python script για εύρεση εννοιών ελληνικών λέξεων
# από το Τριανταφυλλίδης Οn-Line Λεξικό της κοινής νεοελληνικής
# http://www.komvos.edu.gr/dictionaries/dictonline/DictOnLineTri.htm
#
# Version: 0.3 (alpha)
# By Raoul (ok this is a nickname)
# Άδεια: GPL (http://www.gnu.org/licenses/gpl.html)
# ABSOLUTELY NO WARRANTY
#
# Έχει δύο χρήσεις (modes):
#
# 1: ./lexiko.py λέξη
# 	Σ' αυτο το mode επιστρέφει τις έννοιες της λέξης στο terminal.
#
# 2: ./lexiko.py -d
#	Σ'αυτό το mode λειτουργεί ως υποτυπώδης dictionary server, ώστε
#	να μπορούν να αναζητηθούν έννοιες μέσω ενός dictionary client, όπως
#	το gnome-dictionary. Ο server τρέχει στο loopback interface ή
#	(127.0.0.1) στο port 2628. Δεν υποστηρίζει threads, οπότε
#	ένα πιθανό 2ο ταυτόχρονο connection μπαίνει στην ουρά. Σταματάει με Ctrl-C
#	Μερικές παράμετροι μπορούν να οριστούν πιο κάτω.
#
#	ΠΡΟΣΟΧΗ: το server mode είναι σε πρώιμο στάδιο και υπάρχει μόνο για
#	λόγους testing, οπότε:
#		1. Μην τρέχετε το server mode ως root
#		2. Τρέξτε το μόνο τοπικά, χωρίς να επιτρέπετε συνδέσεις
#		μέσω δικτύου.
#
#	Εντολές που υποστηρίζονται:
#	CLIENT, DEFINE, QUIT
#	Για περισσότερες πληροφορίες: ftp://ftp.isi.edu/in-notes/rfc2229.txt
#
#	Παρακάτω είναι ενα δείγμα υποτυπώδους επικοινωνίας server-client
#	για τεστ μέσω telnet:
#
#	220 Welcome message
#	CLIENT "my_dict_client 2.3.14"
#	250 ok
#	DEFINE * "mialeksi"
#	150 2 definitions retrieved
#	151 ORISMOI...
#	...ORISMOI
#	151 ORISMOI
#	.
#	250 ok
#	QUIT
#	221 bye
#

# Μερικές παράμετροι (μόνο για server mode)
Host = 'localhost'	# Το hostname ή IP του interface στο οποίο τρέχει ο σερβερ
Port = 2628			# Το listening port (Αυτό είναι το default του Dict Protocol Server)
Buf = 1024			# Το receive buffer (Αυτό είναι το default του Dict Protocol Server)
MaxConn = 5			# Ο μέγιστος αριθμός των _queued_ συνδέσεων στο σερβερ
Print_Log = True	# Print server log messages to stdout
Print_Client_Com = True		# Print the client commands to stdout
Wrap_At_Column = 60		# Text Wrap των εννοιών σ'αυτή τη στήλη


import sys
import socket
import re
import httplib
import textwrap


def Start_Server():
	""" Εδώ ξεκινάει ο server. Για σερβερ με sockets (από τα socket module examples):
	A server must perform the sequence:	socket(), bind(), listen(), accept()
	accept() should be repeated to serve more than one client.
	"""
	# Δημιουργία του TCP socket
	TCPsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	# Enable the SO_REUSEADDR socket option. (Απαραίτητο, ώστε να μπορεί να ξαναχρησιμοποιηθεί το
	# network interface και να μην δίνει error 'Address already in use' σε επανεκίνηση του σερβερ.
	TCPsock.setsockopt( socket.SOL_SOCKET, socket.SO_REUSEADDR, 1 )
	# Bind & listen
	TCPsock.bind((Host, Port))
	TCPsock.listen(MaxConn)
	
	# Prints the "server started" message
	Srv_Log_Msg(100,('',''))
	
	try:
		# Εδώ είναι το loop του σερβερ για να δεχτεί συνδέσεις
		while True:
			# Accept connections
			# Conn: το νέο socket που δημιουργείται όταν συνδέεται ο client και μέσω 
			# του οποίου γίνεται η επικοινωνία.
			# Addr: ένα tuple με τα IP και local port του client ('127.0.0.1', 39444)
			Conn, Addr = TCPsock.accept()
			
			# Βάλε timeout (sec) στο νέο socket (Conn)
			#Conn.settimeout(30.0)
			
			# Client connected  Print msg
			Srv_Log_Msg(201, Addr)
			
			# Start talking to client
			Dict_Proto_Server(Conn, Addr)
			
			# Client disconnected. Print msg και Close new socket (Conn)
			Srv_Log_Msg(202, Addr)
			Conn.close()
	except KeyboardInterrupt:
		Srv_Log_Msg(101,('',''))
		TCPsock.close()
		sys.exit(0)


def Dict_Proto_Server(ConnSock, Addr):
	""" Dictionary Protocol Server
	"""
	# Supported Client Commands
	# To dictionary exei  command : regexp gia elegxo tou command se kapoia alli synartisi
	# ((NOTE)telika mallon den xreiazetai)
	Client_Commands = {
		'CLIENT' : '^CLIENT "([^"]+)"\r\n$',
		'DEFINE' : '^DEFINE . "([^"]+)"\r\n$',
		'QUIT' : 'QUIT\r\n'
		}
	
	# Not Implemented Client Commands
	Not_Implemented_Client_Commands = ['MATCH', 'SHOW', 'STATUS', 'HELP', 'OPTION', 'AUTH', 'SASLAUTH']
	
	# Server Respnoses ( 150 and 151 codes (Text Responses) are constructed separately)
	Server_Responses = {
		'WELCOME' : '220 Yo <> Welcome\r\n',
		'COMMAND_COMPLETE' : '250 ok\r\n',
		'CLOSE_CONNECTION' : '221 bye\r\n',
		'NO_MATCH' : '552 No match\r\n',
		'NOT_IMPLEMENTED' : '502 Command not implemented\r\n',
		'NOT_RECOGNIZED' : '500 Syntax error, command not recognized\r\n',
		'ILLEGAL_PARAM' : '501 Syntax error, illegal parameters\r\n'
		}
	
	# Στείλε το welcome msg
	ConnSock.send(Server_Responses['WELCOME'])
	Srv_Log_Msg(401, Addr)
	
	# Εδώ είναι το loop της επικοινωνίας με τον client
	while True:
		# Receive data from clients
		Data = ConnSock.recv(Buf)
		
		# If Print_Log is True, print the client commands
		if Print_Client_Com:print "Received command: ", repr(Data)
		
		# If the client sends blank data or the QUIT command, then this is the end
		if (( not Data ) or ( Data.upper() == Client_Commands['QUIT'] )):
			break
		else:
			# Απομόνωση της εντολής. Σβήνονται και τα \r\n, ώστε αν ο client δώσει
			# πχ MATCH\r\n (που δεν πρόκειται να συμβεί σε νορμαλ κατάσταση), να
			# αναγνωριστεί η εντολή πχ ως not implemented παρακάτω.
			command = Data.split(' ', 1)
			command[0] = re.sub('\r\n$', '', command[0])
			
			# Check if the client command is supported
			if command[0].upper() in Client_Commands:
				# Start processing
				
				# Client sent CLIENT command, no action needed, just send ok
				if command[0].upper() == 'CLIENT':
					ConnSock.send(Server_Responses['COMMAND_COMPLETE'])
					Srv_Log_Msg(402, Addr)
				
				# Client sent DEFINE command
				# Δεν υποστηρίζονται παράμετροι (* ή ! κτλ) στο DEFINE, οπότε δεν γίνεται καμία άλλη
				# επεξεργασία της εντολής εκτός από την απομόνωση της ζητούμενης λέξης 
				# στο function Process_Client_Request
				elif command[0].upper() == 'DEFINE':
					# This is where we get the difinitions
					Srv_Log_Msg(403, Addr)
					
					Server_Text_Response = Process_Client_Request(Data, Client_Commands['DEFINE'], Addr)
					
					# Check if there were any errors
					if Server_Text_Response in Server_Responses:
						# Send the returned error response to the client
						ConnSock.send(Server_Responses[Server_Text_Response])
					else:
						# Ola kala
						
						# Server Response 150 (Number of definitions).
						# Actually, this is the number of the sources of all the definitions.
						# Only the Komvos is used as a resource, so this is 1 for now (TODO)
						Number_Of_Definition_Sources = 1
						ConnSock.send('150 ' + str(Number_Of_Definition_Sources) + ' definitions retrieved\r\n')
						
						# Send The definitions
						ConnSock.send(Server_Text_Response)
						
						# Right after the definitions, the server must send the 250 response
						ConnSock.send(Server_Responses['COMMAND_COMPLETE'])
						Srv_Log_Msg(404, Addr)
				
			# Check if the client command is not implemented
			elif command[0].upper() in Not_Implemented_Client_Commands:
				ConnSock.send(Server_Responses['NOT_IMPLEMENTED'])
				Srv_Log_Msg(501, Addr)
			
			# Assumes that the client command is not recognized
			else:
				ConnSock.send(Server_Responses['NOT_RECOGNIZED'])
				Srv_Log_Msg(502, Addr)
	
	# Send goodbye msg
	ConnSock.send(Server_Responses['CLOSE_CONNECTION'])


def pseudo_detect_encoding(user_input):
	"""Μαϊμουδο-ανίχνευση του encoding της λέξης με τη λογική:
	αν δεν είναι utf-8, τότε σίγουρα είναι iso-8859-7.
	Δες: sys.stdin.encoding και το module locale
	Χρειάζεται να εκτελεστεί _μία_ φορα. Το my_encoding είναι στο global table.
	"""
	global my_encoding
	try:
		user_input = user_input.decode('utf-8')
		my_encoding = "utf-8"
	except UnicodeDecodeError:
		my_encoding = "iso-8859-7"


def Process_Client_Request(Data, Pattern, Addr):
	"""Process request
	"""
	# Check the user encoding
	pseudo_detect_encoding(Data)
	
	# Decode and find word
	Data = Data.decode(my_encoding)
	the_word = []
	n = re.compile(Pattern, re.IGNORECASE)
	the_word = n.findall(Data)
	
	# Send the list (the_word) to some tests
	if not Do_Security_Checks(the_word, Addr):
		Srv_Log_Msg(503, Addr)
		return 'ILLEGAL_PARAM'
	
	# Get the Greek definitions from Komvos
	GR_definitions = get_definitions_komvos(the_word[0])
	
	# Return a NO_MATCH response if no definitions are found
	if len(GR_definitions) == 0:
		Srv_Log_Msg(500, Addr)
		return 'NO_MATCH'
	
	# Construct proper Dictionary Server Text Responses
	GR_dict_definitions = Construct_Dict_Definitions(the_word[0], GR_definitions)
	
	Serv_Resp = GR_dict_definitions.encode(my_encoding)
	
	return Serv_Resp


def Do_Security_Checks(the_word, Addr):
	"""Some security(?) checks.
	Σημαντικό για αποφυγή κάποιας στραβής, επειδή σε server mode
	λειτουργεί στην ουσία και σαν client για web sites...
	"""
	# Test 1: Check if there is a word
	try:
		word = the_word[0]
	except IndexError:
		Srv_Log_Msg(701, Addr)
		return False
	
	# Test 2: Check if there is a space separated series of words (word1 word2) etc
	word = word.strip()
	word = word.split(' ', 1)
	if len(word) <> 1:
		Srv_Log_Msg(702, Addr)
		return False
	
	# Test 3: Check if the word contains any non-alphanumeric or numeric characters
	n = re.compile('[\W\d]', re.UNICODE)
	non_alpha = n.findall(word[0])
	if len(non_alpha) > 0:
		Srv_Log_Msg(703, Addr)
		return False
	
	# Test 4: Check the word length
	if len(word[0]) > 25:
		Srv_Log_Msg(704, Addr)
		return False
	
	return True


def get_my_results(the_url, word_or_id):
	"""Επιστρέφει τα περιεχόμενα της web page
	"""
	myurl = the_url + word_or_id
	conn = httplib.HTTPConnection("www.komvos.edu.gr")
	conn.request("POST", myurl)
	r1 = conn.getresponse()
	result1 = r1.read()
	return result1


def get_definitions_komvos(the_word):
	"""Retrieves the greek definitions of the requested word
	from www.komvos.edu.gr and returns a list.
	"""
	# Get a list of the word IDs (mycodes) from komvos
	the_word = the_word.encode('iso-8859-7')
	data1 = get_my_results("/dictonlineplsql/simple_search.search?target_dict=1&the_word=", the_word)
	data1 = data1.decode('iso-8859-7')
	mycodes = re.findall('<option value="([\d]+)"', data1)
	
	# Gets the definition for each word ID
	# and returns a list (definitions) with the definitions
	definitions = []
	for myid in mycodes:
		data1 = get_my_results("/dictonlineplsql/simple_search.display_full_lemma?target_dict=1&the_lemma_id=", myid)
		data1 = data1.decode('iso-8859-7')
		# Isolates the definition from the rest of the page
		aaa = re.findall(r'<font class=lemma>\n(.*)</font>', data1)
		# Strips HTML tags from the definition
		bbb = re.sub('<[^>]+>', '', aaa[0])
		definitions.append(bbb)
	
	return definitions


def Construct_Dict_Definitions(the_Word, the_Definitions):
	"""Constructs proper dict server 151 text responses (text with the definitions).
	"""
	# Contains the word and the source of definitions
	Serv_Resp = u'151 "' + the_Word + u'" gcide "Τριανταφυλλίδης Οn-Line Λεξικό της κοινής νεοελληνικής"\r\n'
	# Wrap text and end each line with CRLF
	for num, definition in enumerate(the_Definitions):
		Serv_Resp = Serv_Resp + str(num + 1) + u'. '
		Definition_Wrapped = textwrap.wrap(definition, Wrap_At_Column)
		for line in Definition_Wrapped:
			Serv_Resp = Serv_Resp + u'   ' + line + u'\r\n'
		Serv_Resp = Serv_Resp + u'\r\n'
	# The last line must contain only a period (.) and end with CRLF
	Serv_Resp = Serv_Resp + u'.\r\n'
	
	return Serv_Resp


def Srv_Log_Msg(Code, Client):
	"""Prints server messages. These are custom codes (koinws paparies), not related to the
	dictionary protocol server response codes.
	"""
	Addr = str(Client[0]) + ':' + str(Client[1])
	
	Server_Log_Messages = {
		100 : 'Dictionary Server started on interface: ' + Host + '  port: ' + str(Port),
		101 : 'Dictionary Server shutting down...',
		201 : 'Client from ' + Addr + ' connected',
		202 : 'Client from ' + Addr + ' disconnected',
		401 : 'Sent welcome message to client at ' + Addr,
		402 : 'Client from ' + Addr + ' introduced itself',
		403 : 'Client from ' + Addr + ' requested definitions',
		404 : 'Sent definitions to client at ' + Addr,
		500 : 'No definitions found for client at ' + Addr,
		501 : 'Client from ' + Addr + ' issued not implemented command',
		502 : 'Client from ' + Addr + ' issued not recognized command',
		503 : 'Client from ' + Addr + ' used illegal parameters',
		701 : 'Request from client ' + Addr + ' failed in Test 1: NO WORD',
		702 : 'Request from client ' + Addr + ' failed in Test 2: NOT SINGLE WORD',
		703 : 'Request from client ' + Addr + ' failed in Test 3: NOT ALPHA',
		704 : 'Request from client ' + Addr + ' failed in Test 4: TOO LENGTHY'
		}
	
	if Print_Log:
		# Print ta server messages
		print 'Log: ' + Server_Log_Messages[Code]
	elif str(Code)[0:1] == '1':
		# Αλλιώς δώσε μόνο "server started" & "server stopped"
		print Server_Log_Messages[Code]


def usage():
	print """
Usage: there are 2 modes:
Mode 1:  ./lexiko.py WORD
  The definitions are printed to stdout.
Mode 2:  ./lexiko.py -d
  In this mode the script acts as a basic Dictionary Server
  for use with dict clients, like gnome-dictionary.
  By default, the server runs on the loopback interface.
  Do not run this script as root. This mode exists only for testing.	
"""


def Run_Only_Cli_Lexiko():
	"""This is mode 1. Επιστρέφει τα definitions στο stdout
	"""
	Responses = {
		1 : u'Αναζήτηση των εννοιών...\n',
		2 : u'Δεν βρέθηκαν εννοιες\n',
		3 : u'Πρόβλημα'
		}
	
	my_word = sys.argv[1]
	pseudo_detect_encoding(my_word)
	print Responses[1].encode(my_encoding)
	my_word = my_word.decode(my_encoding)
	GR_Definition_List = get_definitions_komvos(my_word)
	if len(GR_Definition_List) == 0:
		print Responses[2].encode(my_encoding)
	for definition in GR_Definition_List:
		print definition + '\n'


def Parse_Cli_Arguments():
	if len(sys.argv) <> 2:
		usage()
	elif sys.argv[1] == '-h':
		usage()
	elif sys.argv[1] == '-d':
		Start_Server()
	elif sys.argv[1][0:1] == '-':
		usage()
	else:
		Run_Only_Cli_Lexiko()


# Εδώ ξεκινάει
if __name__=='__main__':
	Parse_Cli_Arguments()

